#region usings
using System;
using System.ComponentModel.Composition;

using VVVV.PluginInterfaces.V1;
using VVVV.PluginInterfaces.V2;
using VVVV.Utils.VColor;
using VVVV.Utils.VMath;

using VVVV.Core.Logging;
#endregion usings

namespace VVVV.Nodes
{
	public struct Exhibit
	{
		public RGBAColor Color;
		public Spread<string> Words;
		public Matrix4x4 Transformation;
		
		public Spread<Matrix4x4> GetTextTransforms()
		{
			int wordCount = Words.SliceCount;
			
			var ret = new Spread<Matrix4x4>(wordCount);
			
			for (int i = 0; i < wordCount; i++)
			{
				ret[i] = VMath.Translate(0, i * -0.8, 0) * Transformation; 
			}
			
			return ret;
		}
	}
	
	[PluginInfo(Name = "Join", Category = "Exhibit")]
	public class JoinExhibit : IPluginEvaluate
	{
	
		[Input("Color")]
		ISpread<RGBAColor> FColorIn;
		
		[Input("Words")]
		ISpread<ISpread<string>> FWordsIn;
		
		[Output("Output")]
		ISpread<Exhibit> FOutput;
		
		public void Evaluate(int SpreadMax)
		{
			int spreadMax = FColorIn.CombineWith(FWordsIn);	
			
			FOutput.SliceCount = spreadMax;
		
			for (int i = 0; i < spreadMax; i++)
			{
				var newExhibit = new Exhibit();
				
				newExhibit.Color = FColorIn[i];
				
				var wordCount = FWordsIn[i].SliceCount;
				
				newExhibit.Words = new Spread<string>(wordCount);
				
				for (int j = 0; j < wordCount; j++)
				{
					 newExhibit.Words[j] = FWordsIn[i][j];
				}
				
				FOutput[i] = newExhibit;
			}
		}
	}
	
	[PluginInfo(Name = "Split", Category = "Exhibit")]
	public class SplitExhibit : IPluginEvaluate
	{
	
		[Input("Input")]
		ISpread<Exhibit> FInput;
		
		[Output("Color")]
		ISpread<RGBAColor> FColorOut;
		
		[Output("Words")]
		ISpread<ISpread<string>> FWordsOut;
		
		[Output("Word Transforms")]
		ISpread<ISpread<Matrix4x4>> FWordsTransOut;
		
		[Output("Transform")]
		ISpread<Matrix4x4> FTransformOut;
		
		public void Evaluate(int SpreadMax)
		{
			FColorOut.SliceCount = SpreadMax;
			FWordsOut.SliceCount = SpreadMax;
			FWordsTransOut.SliceCount = SpreadMax;
			FTransformOut.SliceCount = SpreadMax;
			
			for (int i = 0; i < SpreadMax; i++)
			{
				FColorOut[i] = FInput[i].Color;
				FTransformOut[i] = FInput[i].Transformation;
				
				int wordCount = FInput[i].Words.SliceCount;
				FWordsOut[i].SliceCount = wordCount;
				
				for (int j = 0; j < wordCount; j++)
				{
					FWordsOut[i][j] = FInput[i].Words[j];
				}
				
				FWordsTransOut[i] = FInput[i].GetTextTransforms();
			}
		}
	}
	
	
	[PluginInfo(Name = "SetTransform", Category = "Exhibit")]
	public class SetTransformExhibit : IPluginEvaluate
	{
		
	
		[Input("Input")]
		ISpread<Exhibit> FInput;
		
		[Input("Transform")]
		ISpread<Matrix4x4> FTransformIn;
		
		[Output("Output")]
		ISpread<Exhibit> FOutput;
		
		public void Evaluate(int SpreadMax)
		{
			FOutput.SliceCount = SpreadMax;
			for (int i = 0; i < SpreadMax; i++)
			{
				Exhibit slice = FInput[i];
				slice.Transformation = FTransformIn[i];
				FOutput[i] = slice; 
			}
		}
		
	}

	public abstract class InputMorph<UnknownType> : IPluginEvaluate
	{
		[Input("Switch")]
		ISpread<double> FMorph;
	
		[Input("Input", IsPinGroup = true)]
		ISpread<ISpread<UnknownType>> FInput;

		[Output("Output")]
		ISpread<UnknownType> FOutput;

		//called when data for any output pin is requested
		public void Evaluate(int SpreadMax)
		{
			int sliceCount = FMorph.SliceCount;
			
			for (int i = 0; i < FInput.SliceCount; i++)
				sliceCount = sliceCount.CombineWith<UnknownType>(FInput[i]);

			FOutput.SliceCount = sliceCount;
			
			for (int i = 0; i < sliceCount; i++)
			{
				double morph = FMorph[i];
				
				int firstPin = (int)Math.Floor(morph);
				int secondPin = firstPin+1;
				
				FOutput[i] = Morph(
					FInput[firstPin][i],
					FInput[secondPin][i], 
					morph % 1);
			}
		}
		
		protected abstract UnknownType Morph(UnknownType a, UnknownType b, double morph);
	}

	[PluginInfo(Name = "InputMorph", Category = "String")]
	public class InputMorphString : InputMorph<string>
	{
		protected override string Morph(string a, string b, double morph)
		{
			
			int la = a.Length;
			int lb = b.Length;
			int wordLength = (int)Math.Round(VMath.Lerp(la, lb, morph));
			
			char[] chars = new char[wordLength];
			
			for(int i=0; i<wordLength; i++)
			{
				var ca = (int)a[i%la];
				var cb = (int)b[i%lb];
				
				var cout = (int) Math.Round(VMath.Lerp(ca, cb, morph));
				
				chars[i] = (char)cout;
			}
			
			return new string(chars);
		}
	}
	
	
}
